import os
import random
import shutil
from bisect import bisect_right
from subprocess import PIPE, call
from typing import TYPE_CHECKING, Literal, Optional, Union, cast

from ape.api import (
    BlockAPI,
    ForkedNetworkAPI,
    PluginConfig,
    ReceiptAPI,
    SubprocessProvider,
    TestProviderAPI,
    TraceAPI,
    TransactionAPI,
)
from ape.exceptions import (
    ContractLogicError,
    OutOfGasError,
    SubprocessError,
    TransactionError,
    VirtualMachineError,
)
from ape.logging import logger
from ape.utils import cached_property
from ape_ethereum.provider import Web3Provider
from ape_test import ApeTestConfig
from eth_pydantic_types import HexBytes, HexBytes32
from eth_typing import HexStr
from eth_utils import add_0x_prefix, is_0x_prefixed, is_hex, to_hex
from pydantic import field_validator, model_validator
from pydantic_settings import SettingsConfigDict
from web3 import HTTPProvider, Web3
from web3.exceptions import ContractCustomError
from web3.exceptions import ContractLogicError as Web3ContractLogicError
from web3.exceptions import ExtraDataLengthError
from web3.gas_strategies.rpc import rpc_gas_price_strategy

try:
    from web3.middleware import ExtraDataToPOAMiddleware  # type: ignore
except ImportError:
    from web3.middleware import geth_poa_middleware as ExtraDataToPOAMiddleware  # type: ignore

from web3.middleware.validation import MAX_EXTRADATA_LENGTH
from yarl import URL

from ape_foundry.constants import EVM_VERSION_BY_NETWORK
from ape_foundry.exceptions import (
    FoundryNotInstalledError,
    FoundryProviderError,
    FoundrySubprocessError,
)
from ape_foundry.trace import AnvilTransactionTrace

try:
    from ape_optimism import Optimism  # type: ignore
except ImportError:
    Optimism = None  # type: ignore

if TYPE_CHECKING:
    from ape.types import AddressType, BlockID, ContractCode, SnapshotID


EPHEMERAL_PORTS_START = 49152
EPHEMERAL_PORTS_END = 60999
DEFAULT_PORT = 8545
FOUNDRY_CHAIN_ID = 31337


class FoundryForkConfig(PluginConfig):
    upstream_provider: Optional[str] = None
    block_number: Optional[int] = None
    evm_version: Optional[str] = None


class FoundryNetworkConfig(PluginConfig):
    host: Optional[Union[str, Literal["auto"]]] = None
    """The host address or ``"auto"`` to use localhost with a random port (with attempts)."""

    manage_process: bool = True
    """
    If ``True`` and the host is local and Anvil is not running, will attempt to start.
    Defaults to ``True``. If ``host`` is remote, will not be able to start.
    """

    evm_version: Optional[str] = None
    """The EVM hardfork to use, e.g. `shanghai`."""

    # Retry strategy configs, try increasing these if you're getting FoundrySubprocessError
    request_timeout: int = 30
    fork_request_timeout: int = 300
    process_attempts: int = 5

    # RPC defaults
    base_fee: int = 0
    priority_fee: int = 0
    gas_price: int = 0

    # For setting the values in --fork and --fork-block-number command arguments.
    # Used only in FoundryForkProvider.
    # Mapping of ecosystem_name => network_name => FoundryForkConfig
    fork: dict[str, dict[str, FoundryForkConfig]] = {}

    disable_block_gas_limit: bool = False
    """
    Disable the ``call.gas_limit <= block.gas_limit`` constraint.
    """

    auto_mine: bool = True
    """
    Automatically mine blocks instead of manually doing so.
    """

    block_time: Optional[int] = None
    """
    Set a block time to allow mining to happen on an interval
    rather than only when a new transaction is submitted.
    """

    use_optimism: Optional[bool] = None
    """
    Configure the node to run with the `--optimism` flag.
    NOTE: When using Optimism-based networks (including Base),
    this flag is automatically added.
    """

    model_config = SettingsConfigDict(extra="allow")

    @field_validator("fork", mode="before")
    @classmethod
    def _validate_fork(cls, value):
        return value or {}


def _call(*args):
    return call([*args], stderr=PIPE, stdout=PIPE, stdin=PIPE)


class FoundryProvider(SubprocessProvider, Web3Provider, TestProviderAPI):
    _host: Optional[str] = None
    attempted_ports: list[int] = []
    cached_chain_id: Optional[int] = None
    _did_warn_wrong_node = False
    _disconnected: Optional[bool] = None

    @property
    def unlocked_accounts(self) -> list["AddressType"]:
        return list(self.account_manager.test_accounts._impersonated_accounts)

    @property
    def mnemonic(self) -> str:
        return self._test_config.mnemonic

    @property
    def number_of_accounts(self) -> int:
        return self._test_config.number_of_accounts

    @property
    def initial_balance(self) -> int:
        """
        Have to convert the WEI value to ETH, like Anvil expects.
        """
        bal_in_wei = self._test_config.balance
        return bal_in_wei // 10**18

    @property
    def process_name(self) -> str:
        return "anvil"

    @property
    def connection_id(self) -> Optional[str]:
        return f"{self.network_choice}:{self._host}"

    @property
    def timeout(self) -> int:
        return self.settings.request_timeout

    @property
    def _clean_uri(self) -> str:
        try:
            return str(URL(self.uri).with_user(None).with_password(None))
        except ValueError:
            # Likely isn't a real URI.
            return self.uri

    @property
    def use_optimism(self) -> bool:
        return self.settings.use_optimism or (
            self.settings.use_optimism is not False
            and Optimism is not None
            and isinstance(self.network.ecosystem, Optimism)
        )

    @property
    def _port(self) -> Optional[int]:
        return URL(self.uri).port

    @property
    def chain_id(self) -> int:
        if self.cached_chain_id is not None:
            return self.cached_chain_id

        elif self.cached_chain_id is None and self._web3 is not None and hasattr(self.web3, "eth"):
            self.cached_chain_id = self.web3.eth.chain_id
            return self.cached_chain_id

        else:
            return FOUNDRY_CHAIN_ID

    @cached_property
    def anvil_bin(self) -> str:
        anvil = shutil.which("anvil")

        if not anvil:
            raise FoundryNotInstalledError()
        elif _call(anvil, "--version") != 0:
            raise FoundrySubprocessError(
                "Anvil executable returned error code. See ape-foundry README for install steps."
            )

        return anvil

    @property
    def uri(self) -> str:
        if self._host is not None:
            return self._host

        elif config_host := self.settings.host:
            if config_host == "auto":
                self._host = "auto"
                return self._host
            if not config_host.startswith("http"):
                if "127.0.0.1" in config_host or "localhost" in config_host:
                    self._host = f"http://{config_host}"
                else:
                    self._host = f"https://{config_host}"
            else:
                self._host = config_host

            if "127.0.0.1" in config_host or "localhost" in config_host:
                host_without_http = self._host[7:]
                if ":" not in host_without_http:
                    self._host = f"{self._host}:{DEFAULT_PORT}"

        else:
            self._host = f"http://127.0.0.1:{DEFAULT_PORT}"

        return self._host

    @property
    def http_uri(self) -> str:
        # NOTE: Overriding `Web3Provider.http_uri` implementation
        return self.uri

    @property
    def ws_uri(self) -> str:
        # NOTE: Overriding `Web3Provider.ws_uri` implementation
        return "ws" + self.uri[4:]  # Remove `http` in default URI w/ `ws`

    @property
    def is_connected(self) -> bool:
        if self._disconnected is True or self._host in ("auto", None):
            # Hasn't tried yet.
            return False

        self._set_web3()
        return self._web3 is not None

    @property
    def gas_price(self) -> int:
        if self.process is not None:
            # NOTE: Workaround for bug where RPC does not honor CLI flag.
            return self.settings.gas_price

        # Not managing node so must use RPC.
        return self.web3.eth.gas_price

    @cached_property
    def _test_config(self) -> ApeTestConfig:
        return cast(ApeTestConfig, self.config_manager.get_config("test"))

    @property
    def auto_mine(self) -> bool:
        return self.make_request("anvil_getAutomine", [])

    @auto_mine.setter
    def auto_mine(self, value) -> None:
        self.make_request("anvil_setAutomine", [value])

    @property
    def evm_version(self) -> Optional[str]:
        return self.settings.evm_version

    @property
    def settings(self) -> FoundryNetworkConfig:
        return cast(FoundryNetworkConfig, super().settings)

    def connect(self):
        """
        Start the foundry process and verify it's up and accepting connections.
        **NOTE**: Must set port before calling 'super().connect()'.
        """
        self._disconnected = False
        if "APE_FOUNDRY_HOST" in os.environ:
            self._host = os.environ["APE_FOUNDRY_HOST"]

        elif self._host is None:
            self._host = self.uri

        if self.is_connected:
            # Connects to already running process
            self._start()

        elif self.settings.manage_process and (
            "localhost" in self._host or "127.0.0.1" in self._host or self._host == "auto"
        ):
            # Only do base-process setup if not connecting to already-running process
            # and is running on localhost.
            super().connect()
            if self._host:
                self._set_web3()
                if not self._web3:
                    # Process attempts to get started at this point.
                    self._start()
                    if not self.stdout_logs_path.is_file():
                        # Process output not being captured
                        return

                    wait_for_key = "Listening on"
                    timeout = 10
                    iterations = 0
                    while iterations < timeout:
                        logged_lines = [x for x in self.stdout_logs_path.read_text().split("\n")]
                        for line in logged_lines:
                            if line.startswith(wait_for_key):
                                return

                        iterations += 1
                        if iterations == timeout:
                            raise FoundryProviderError(
                                "Timed-out waiting for process to begin listening."
                            )

                else:
                    # The user configured a host and the anvil process was already running.
                    logger.info(
                        f"Connecting to existing '{self.process_name}' "
                        f"at host '{self._clean_uri}'."
                    )
            else:
                for _ in range(self.settings.process_attempts):
                    try:
                        self._start()
                        break
                    except FoundryNotInstalledError:
                        # Is a sub-class of `FoundrySubprocessError` but we to still raise
                        # so we don't keep retrying.
                        raise
                    except SubprocessError as exc:
                        logger.info("Retrying anvil subprocess startup: %r", exc)
                        self._host = None
        else:
            raise FoundryProviderError(
                f"Failed to connect to remote Anvil node at '{self._clean_uri}'."
            )

    def _set_web3(self):
        if not self._host:
            return

        self._web3 = Web3(HTTPProvider(self.uri, request_kwargs={"timeout": self.timeout}))

        try:
            is_connected = self._web3.is_connected()
        except Exception:
            is_connected = False

        if not is_connected:
            self._web3 = None
            return

        # Verify is actually a Foundry provider,
        # or else skip it to possibly try another port.
        client_version = self._web3.client_version.lower()
        if "anvil" in client_version:
            self._web3.eth.set_gas_price_strategy(rpc_gas_price_strategy)
        elif self._port is not None:
            raise FoundryProviderError(
                f"Port '{self._port}' already in use by another process that isn't an Anvil node."
            )
        else:
            # Not sure if possible to get here.
            raise FoundryProviderError("Failed to start Anvil process.")

        def check_poa(block_id) -> bool:
            try:
                block = self.web3.eth.get_block(block_id)
            except ExtraDataLengthError:
                return True
            else:
                return (
                    "proofOfAuthorityData" in block
                    or len(block.get("extraData", "")) > MAX_EXTRADATA_LENGTH
                )

        # Handle if using PoA
        if any(map(check_poa, (0, "latest"))):
            self._web3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

    def _start(self):
        if self.is_connected:
            return

        use_random_port = self._host == "auto"
        if use_random_port:
            self._host = None

            if DEFAULT_PORT not in self.attempted_ports:
                # First, attempt the default port before anything else.
                self._host = f"127.0.0.1:{DEFAULT_PORT}"

            # Pick a random port
            port = random.randint(EPHEMERAL_PORTS_START, EPHEMERAL_PORTS_END)
            max_attempts = 25
            attempts = 0
            while port in self.attempted_ports:
                port = random.randint(EPHEMERAL_PORTS_START, EPHEMERAL_PORTS_END)
                attempts += 1
                if attempts == max_attempts:
                    ports_str = ", ".join([str(p) for p in self.attempted_ports])
                    raise FoundryProviderError(
                        f"Unable to find an available port. Ports tried: {ports_str}"
                    )

            self.attempted_ports.append(port)
            self._host = f"http://127.0.0.1:{port}"

        elif self._host is not None and ":" in self._host and self._port is not None:
            # Append the one and only port to the attempted ports list, for honest keeping.
            self.attempted_ports.append(self._port)

        else:
            self._host = f"http://127.0.0.1:{DEFAULT_PORT}"

        if "127.0.0.1" in self._host or "localhost" in self._host:
            # Start local process
            self.start()

        elif not self.is_connected:
            raise FoundryProviderError(f"Failed to connect to Anvil node at '{self._clean_uri}'.")

    def disconnect(self):
        self._web3 = None
        self._host = None
        super().disconnect()
        self._disconnected = True

    def build_command(self) -> list[str]:
        cmd = [
            self.anvil_bin,
            "--port",
            f"{self._port or DEFAULT_PORT}",
            "--mnemonic",
            self.mnemonic,
            "--accounts",
            f"{self.number_of_accounts}",
            "--derivation-path",
            f"{self.test_config.hd_path}",
            "--balance",
            f"{self.initial_balance}",
            "--steps-tracing",
            "--block-base-fee-per-gas",
            f"{self.settings.base_fee}",
            "--gas-price",
            f"{self.settings.gas_price}",
        ]

        if not self.settings.auto_mine:
            cmd.append("--no-mining")

        if self.settings.block_time is not None:
            cmd.extend(("--block-time", f"{self.settings.block_time}"))

        if self.settings.disable_block_gas_limit:
            cmd.append("--disable-block-gas-limit")

        if evm_version := self.evm_version:
            cmd.extend(("--hardfork", evm_version))

        # Optimism-based networks are different; Anvil provides a flag to make
        # testing more like the real network(s).
        if self.use_optimism:
            cmd.append("--optimism")

        return cmd

    def set_balance(self, account: "AddressType", amount: Union[int, float, str, bytes]):
        is_str = isinstance(amount, str)
        is_key_word = is_str and " " in amount  # type: ignore
        _is_hex = is_str and not is_key_word and amount.startswith("0x")  # type: ignore

        if is_key_word:
            # This allows values such as "1000 ETH".
            amount = self.conversion_manager.convert(amount, int)
            is_str = False

        amount_hex_str: str
        if is_str and not _is_hex:
            amount_hex_str = to_hex(int(amount))
        elif isinstance(amount, (int, bytes)):
            amount_hex_str = to_hex(amount)
        else:
            amount_hex_str = str(amount)

        self.make_request("anvil_setBalance", [account, amount_hex_str])

    def set_timestamp(self, new_timestamp: int):
        self.make_request("evm_setNextBlockTimestamp", [new_timestamp])

    def mine(self, num_blocks: int = 1):
        # NOTE: Request fails when given numbers with any left padded 0s.
        num_blocks_arg = f"0x{HexBytes(num_blocks).hex().replace('0x', '').lstrip('0')}"
        self.make_request("anvil_mine", [num_blocks_arg])

    def snapshot(self) -> str:
        return self.make_request("evm_snapshot", [])

    def restore(self, snapshot_id: "SnapshotID") -> bool:
        snapshot_id = to_hex(snapshot_id) if isinstance(snapshot_id, int) else snapshot_id
        result = self.make_request("evm_revert", [snapshot_id])
        return result is True

    def unlock_account(self, address: "AddressType") -> bool:
        self.make_request("anvil_impersonateAccount", [address])
        return True

    def relock_account(self, address: "AddressType"):
        self.make_request("anvil_stopImpersonatingAccount", [address])

    def get_balance(self, address: "AddressType", block_id: Optional["BlockID"] = None) -> int:
        if result := self.make_request("eth_getBalance", [address, block_id]):
            return int(result, 16) if isinstance(result, str) else result

        raise FoundryProviderError(f"Failed to get balance for account '{address}'.")

    def get_transaction_trace(self, transaction_hash: str, **kwargs) -> TraceAPI:
        return _get_transaction_trace(transaction_hash, **kwargs)

    def get_virtual_machine_error(self, exception: Exception, **kwargs) -> VirtualMachineError:
        if not exception.args:
            return VirtualMachineError(base_err=exception, **kwargs)

        err_data = exception.args[0]

        # Determine message based on the type of error data
        if isinstance(err_data, dict):
            message = str(err_data.get("message", f"{err_data}"))
        else:
            message = err_data if isinstance(err_data, str) else getattr(exception, "message", "")

        if not message:
            return VirtualMachineError(base_err=exception, **kwargs)

        # Handle specific cases based on message content
        foundry_prefix = (
            "Error: VM Exception while processing transaction: reverted with reason string "
        )

        # Handle Foundry error prefix
        if message.startswith(foundry_prefix):
            message = message.replace(foundry_prefix, "").strip("'")
            return self._handle_execution_reverted(exception, message, **kwargs)

        # Handle various cases of transaction reverts
        if "Transaction reverted without a reason string" in message:
            return self._handle_execution_reverted(exception, **kwargs)

        if message.lower() == "execution reverted":
            message = TransactionError.DEFAULT_MESSAGE
            if isinstance(exception, Web3ContractLogicError):
                if custom_msg := self._extract_custom_error(**kwargs):
                    exception.message = custom_msg
            return self._handle_execution_reverted(exception, revert_message=message, **kwargs)

        if "Transaction ran out of gas" in message or "OutOfGas" in message:
            return OutOfGasError(base_err=exception, **kwargs)

        if message.startswith("execution reverted: "):
            message = (
                message.replace("execution reverted: ", "").strip()
                or TransactionError.DEFAULT_MESSAGE
            )
            return self._handle_execution_reverted(exception, revert_message=message, **kwargs)

        # Handle custom errors
        if isinstance(exception, ContractCustomError):
            message = TransactionError.DEFAULT_MESSAGE if message in ("", None, "0x") else message
            return self._handle_execution_reverted(exception, revert_message=message, **kwargs)

        return VirtualMachineError(message, **kwargs)

    # The type ignore is because are using **kwargs rather than repeating.
    def _handle_execution_reverted(  # type: ignore[override]
        self, exception: Exception, revert_message: Optional[str] = None, **kwargs
    ):
        # Assign default message if revert_message is invalid
        if revert_message == "0x":
            revert_message = TransactionError.DEFAULT_MESSAGE
        else:
            revert_message = revert_message or TransactionError.DEFAULT_MESSAGE

        if revert_message.startswith("revert: "):
            revert_message = revert_message[8:]

        # Create and enrich the error
        sub_err = ContractLogicError(base_err=exception, revert_message=revert_message, **kwargs)
        enriched = self.compiler_manager.enrich_error(sub_err)

        # Show call trace if available
        txn = enriched.txn
        if txn and hasattr(txn, "show_trace"):
            if isinstance(txn, TransactionAPI) and txn.receipt:
                txn.receipt.show_trace()
            elif isinstance(txn, ReceiptAPI):
                txn.show_trace()

        return enriched

    # Abstracted for easier testing conditions.
    def _extract_custom_error(self, **kwargs) -> str:
        # Check for custom error.
        trace = None
        if "trace" in kwargs:
            trace = kwargs["trace"]

        elif "txn" in kwargs:
            txn = kwargs["txn"]
            try:
                txn_hash = txn.txn_hash if isinstance(txn.txn_hash, str) else to_hex(txn.txn_hash)
                trace = self.get_transaction_trace(txn_hash)
            except Exception:
                pass

        if trace is not None and (revert_msg := trace.revert_message):
            return revert_msg

        return ""

    def set_block_gas_limit(self, gas_limit: int) -> bool:
        return self.make_request("evm_setBlockGasLimit", [hex(gas_limit)]) is True

    def set_code(self, address: "AddressType", code: "ContractCode") -> bool:
        if isinstance(code, bytes):
            code = to_hex(code)

        elif isinstance(code, str) and not is_0x_prefixed(code):
            code = add_0x_prefix(HexStr(code))

        elif not is_hex(code):
            raise ValueError(f"Value {code} is not convertible to hex")

        self.make_request("anvil_setCode", [address, code])
        return True

    def set_storage(self, address: "AddressType", slot: int, value: HexBytes):
        self.make_request(
            "anvil_setStorageAt",
            [
                address,
                to_hex(HexBytes32.__eth_pydantic_validate__(slot)),
                to_hex(HexBytes32.__eth_pydantic_validate__(value)),
            ],
        )


class FoundryForkProvider(FoundryProvider):
    """
    A Foundry provider that uses ``--fork``, like:
    ``npx foundry node --fork <upstream-provider-url>``.

    Set the ``upstream_provider`` in the ``foundry.fork`` config
    section of your ``ape-config.yaml` file to specify which provider
    to use as your archive node.
    """

    @model_validator(mode="before")
    @classmethod
    def set_upstream_provider(cls, value):
        network = value["network"]
        adhoc_settings = value.get("provider_settings", {}).get("fork", {})
        ecosystem_name = network.ecosystem.name
        plugin_config = cls.config_manager.get_config(value["name"])
        config_settings = plugin_config.get("fork", {})

        def _get_upstream(data: dict) -> Optional[str]:
            return (
                data.get(ecosystem_name, {})
                .get(network.name.replace("-fork", ""), {})
                .get("upstream_provider")
            )

        # If upstream provider set anywhere in provider settings, ignore.
        if name := (_get_upstream(adhoc_settings) or _get_upstream(config_settings)):
            getattr(network.ecosystem.config, network.name).upstream_provider = name

        return value

    @property
    def fork_block_number(self) -> Optional[int]:
        return self._fork_config.block_number

    @property
    def evm_version(self) -> Optional[str]:
        if evm_version := self._fork_config.evm_version:
            return evm_version

        return self.settings.evm_version

    def get_block(self, block_id: "BlockID") -> BlockAPI:
        if isinstance(block_id, str) and block_id.isnumeric():
            block_id = int(block_id)

        block_data = dict(self.web3.eth.get_block(block_id))

        # Fix Foundry-specific differences
        if "baseFeePerGas" in block_data and block_data.get("baseFeePerGas") is None:
            block_data["baseFeePerGas"] = 0

        return self.network.ecosystem.decode_block(block_data)

    def detect_evm_version(self) -> Optional[str]:
        if self.fork_block_number is None:
            return None

        ecosystem = self.forked_network.ecosystem.name
        network = self.forked_network.upstream_network.name
        try:
            hardforks = EVM_VERSION_BY_NETWORK[ecosystem][network]
        except KeyError:
            return None

        keys = sorted(hardforks)
        index = bisect_right(keys, self.fork_block_number) - 1
        return hardforks[keys[index]]

    @property
    def timeout(self) -> int:
        return self.settings.fork_request_timeout

    @property
    def _fork_config(self) -> FoundryForkConfig:
        ecosystem_name = self.network.ecosystem.name
        if ecosystem_name not in self.settings.fork:
            return FoundryForkConfig()  # Just use default

        network_name = self.forked_network.upstream_network.name
        if network_name not in self.settings.fork[ecosystem_name]:
            return FoundryForkConfig()  # Just use default

        return self.settings.fork[ecosystem_name][network_name]

    @property
    def forked_network(self) -> ForkedNetworkAPI:
        return cast(ForkedNetworkAPI, self.network)

    @property
    def upstream_provider_name(self) -> str:
        if upstream_name := self._fork_config.upstream_provider:
            self._fork_config.upstream_provider = upstream_name

        return self.forked_network.upstream_provider.name

    @property
    def fork_url(self) -> str:
        return self.forked_network.upstream_provider.connection_str

    def connect(self):
        super().connect()

        # If using the provider config for upstream_provider,
        # set the network one in this session, so other features work in core.
        with self.forked_network.use_upstream_provider() as upstream_provider:
            try:
                upstream_genesis_block_hash = upstream_provider.get_block(0).hash
            except ExtraDataLengthError as err:
                if isinstance(upstream_provider, Web3Provider):
                    logger.error(
                        f"Upstream provider '{upstream_provider.name}' missing Geth PoA middleware."
                    )
                    upstream_provider.web3.middleware_onion.inject(
                        ExtraDataToPOAMiddleware, layer=0
                    )
                    upstream_genesis_block_hash = upstream_provider.get_block(0).hash
                else:
                    raise FoundryProviderError(f"Unable to get genesis block: {err}.") from err

        if self.get_block(0).hash != upstream_genesis_block_hash:
            logger.warning(
                "Upstream network has mismatching genesis block. "
                "This could be an issue with foundry."
            )

    def build_command(self) -> list[str]:
        if not self.fork_url:
            raise FoundryProviderError("Upstream provider does not have a ``connection_str``.")

        if self.fork_url.replace("localhost", "127.0.0.1").replace("http://", "") == self.uri:
            raise FoundryProviderError(
                "Invalid upstream-fork URL. Can't be same as local anvil node."
            )

        cmd = super().build_command()
        cmd.extend(("--fork-url", self.fork_url))
        if self.fork_block_number is not None:
            cmd.extend(("--fork-block-number", str(self.fork_block_number)))

            if self._fork_config.evm_version is None:
                self._fork_config.evm_version = self.detect_evm_version()

        return cmd

    def reset_fork(self, block_number: Optional[int] = None):
        forking_params: dict[str, Union[str, int]] = {"jsonRpcUrl": self.fork_url}
        block_number = block_number if block_number is not None else self.fork_block_number
        if block_number is not None:
            forking_params["blockNumber"] = block_number

        # # Rest the fork
        result = self.make_request("anvil_reset", [{"forking": forking_params}])
        return result


def _get_transaction_trace(transaction_hash: str, **kwargs) -> TraceAPI:
    # Abstracted for testing purposes.
    return AnvilTransactionTrace(transaction_hash=transaction_hash, **kwargs)



"""
Ape network provider plugin for Foundry (Ethereum development framework and network
implementation written in Node.js).
"""

from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from .provider import FoundryNetworkConfig

    return FoundryNetworkConfig


@plugins.register(plugins.ProviderPlugin)
def providers():
    from ape.api.networks import LOCAL_NETWORK_NAME
    from ape_ethereum.ecosystem import NETWORKS

    from .provider import FoundryForkProvider, FoundryProvider

    yield "ethereum", LOCAL_NETWORK_NAME, FoundryProvider

    for network in NETWORKS:
        yield "ethereum", f"{network}-fork", FoundryForkProvider

    yield "arbitrum", LOCAL_NETWORK_NAME, FoundryProvider
    yield "arbitrum", "mainnet-fork", FoundryForkProvider
    yield "arbitrum", "sepolia-fork", FoundryForkProvider

    yield "avalanche", LOCAL_NETWORK_NAME, FoundryProvider
    yield "avalanche", "mainnet-fork", FoundryForkProvider
    yield "avalanche", "fuji-fork", FoundryForkProvider

    yield "bsc", LOCAL_NETWORK_NAME, FoundryProvider
    yield "bsc", "mainnet-fork", FoundryForkProvider
    yield "bsc", "testnet-fork", FoundryForkProvider

    yield "fantom", LOCAL_NETWORK_NAME, FoundryProvider
    yield "fantom", "opera-fork", FoundryForkProvider
    yield "fantom", "testnet-fork", FoundryForkProvider

    yield "optimism", LOCAL_NETWORK_NAME, FoundryProvider
    yield "optimism", "mainnet-fork", FoundryForkProvider
    yield "optimism", "sepolia-fork", FoundryForkProvider

    yield "polygon", LOCAL_NETWORK_NAME, FoundryProvider
    yield "polygon", "mainnet-fork", FoundryForkProvider
    yield "polygon", "amoy-fork", FoundryForkProvider

    yield "base", LOCAL_NETWORK_NAME, FoundryProvider
    yield "base", "mainnet-fork", FoundryForkProvider
    yield "base", "sepolia-fork", FoundryForkProvider

    yield "blast", LOCAL_NETWORK_NAME, FoundryProvider
    yield "blast", "mainnet-fork", FoundryForkProvider
    yield "blast", "sepolia-fork", FoundryForkProvider


def __getattr__(name: str):
    if name == "FoundryForkProvider":
        from ape_foundry.provider import FoundryForkProvider

        return FoundryForkProvider

    elif name == "FoundryNetworkConfig":
        from ape_foundry.provider import FoundryNetworkConfig

        return FoundryNetworkConfig

    elif name == "FoundryProvider":
        from ape_foundry.provider import FoundryProvider

        return FoundryProvider

    elif name == "FoundryProviderError":
        from ape_foundry.provider import FoundryProviderError

        return FoundryProviderError

    elif name == "FoundrySubprocessError":
        from ape_foundry.provider import FoundrySubprocessError

        return FoundrySubprocessError

    else:
        raise AttributeError(name)


__all__ = [
    "FoundryForkProvider",
    "FoundryNetworkConfig",
    "FoundryProvider",
    "FoundryProviderError",
    "FoundrySubprocessError",
]



import os

import pytest
from ape import convert, reverts
from ape.api import TraceAPI
from ape.api.accounts import ImpersonatedAccount
from ape.contracts import ContractContainer
from ape.exceptions import ContractLogicError, TransactionError, VirtualMachineError
from ape_ethereum.trace import Trace
from ape_ethereum.transactions import TransactionStatusEnum, TransactionType
from eth_pydantic_types import HexBytes32
from eth_utils import to_hex, to_int
from evm_trace import CallType
from hexbytes import HexBytes

from ape_foundry import FoundryProviderError
from ape_foundry.provider import FOUNDRY_CHAIN_ID

TEST_WALLET_ADDRESS = "0xD9b7fdb3FC0A0Aa3A507dCf0976bc23D49a9C7A3"


def test_instantiation(disconnected_provider, name):
    assert disconnected_provider.name == name


def test_connect_and_disconnect(disconnected_provider):
    # Use custom port to prevent connecting to a port used in another test.

    disconnected_provider._host = "http://127.0.0.1:8555"
    disconnected_provider.connect()

    try:
        assert disconnected_provider.is_connected
        assert disconnected_provider.chain_id == FOUNDRY_CHAIN_ID
    finally:
        disconnected_provider.disconnect()

    assert not disconnected_provider.is_connected
    assert disconnected_provider.process is None


def test_gas_price(connected_provider):
    gas_price = connected_provider.gas_price
    assert gas_price == 0


def test_uri_disconnected(disconnected_provider):
    assert disconnected_provider.uri == "http://127.0.0.1:8545"


def test_uri(connected_provider):
    assert connected_provider.uri in connected_provider.uri


def test_set_block_gas_limit(connected_provider):
    gas_limit = connected_provider.get_block("latest").gas_limit
    assert connected_provider.set_block_gas_limit(gas_limit) is True


def test_set_timestamp(connected_provider):
    start_time = connected_provider.get_block("pending").timestamp
    connected_provider.set_timestamp(start_time + 5)
    new_time = connected_provider.get_block("pending").timestamp

    # Adding 5 seconds but seconds can be weird so give it a 1 second margin.
    assert 4 <= new_time - start_time <= 6


def test_mine(connected_provider):
    block_num = connected_provider.get_block("latest").number
    connected_provider.mine(100)
    next_block_num = connected_provider.get_block("latest").number
    assert next_block_num > block_num


def test_restore_failure(connected_provider):
    assert connected_provider.restore(0xFFFF) is False


def test_get_balance(connected_provider, owner):
    assert connected_provider.get_balance(owner.address)


def test_snapshot_and_restore(connected_provider):
    snap = connected_provider.snapshot()

    block_1 = connected_provider.get_block("latest")
    connected_provider.mine()
    block_2 = connected_provider.get_block("latest")
    assert block_2.number > block_1.number
    assert block_1.hash != block_2.hash

    connected_provider.restore(snap)
    block_3 = connected_provider.get_block("latest")
    assert block_1.number == block_3.number
    assert block_1.hash == block_3.hash


@pytest.mark.parametrize(
    "tx_kwargs",
    [
        {},  # NO KWARGS CASE: Should default to type 2
        {"type": 0},
        {"type": 0, "gas_price": 0},
        {"type": 1},
        {"type": 1, "gas_price": 0},
        {"type": 2},
        {"type": 2, "max_priority_fee": 0},
        {"type": 2, "base_fee": 0, "max_priority_fee": 0},
    ],
)
def test_unlock_account(connected_provider, contract_a, accounts, tx_kwargs):
    actual = connected_provider.unlock_account(TEST_WALLET_ADDRESS)
    assert actual is True

    # Tell Anvil we no longer want this unlocked.
    connected_provider.relock_account(TEST_WALLET_ADDRESS)

    # Unlock using the more public API.
    acct = accounts[TEST_WALLET_ADDRESS]
    assert isinstance(acct, ImpersonatedAccount)

    # Ensure can transact.
    # NOTE: Using type 0 to avoid needing to set a balance.
    acct.balance += 1_000_000_000_000_000_000
    receipt_0 = contract_a.methodWithoutArguments(sender=acct, **tx_kwargs)

    # Ensure we can deploy.
    container = ContractContainer(contract_a.contract_type)
    new_contract = container.deploy(acct, acct, sender=acct)
    assert new_contract.address != contract_a.address

    assert not receipt_0.failed


def test_get_transaction_trace(connected_provider, contract_instance, owner):
    receipt = contract_instance.setNumber(10, sender=owner)

    actual = connected_provider.get_transaction_trace(receipt.txn_hash)
    assert isinstance(actual, TraceAPI), f"{type(actual)}"
    assert actual == receipt.trace

    # Indirectly calls `connected_provider.get_transaction_trace()`
    assert isinstance(receipt.trace, TraceAPI)


def test_get_transaction_trace_call_tree(connected_provider, sender, receiver):
    transfer = sender.transfer(receiver, 1)
    trace = connected_provider.get_transaction_trace(transfer.txn_hash)
    assert isinstance(trace, Trace)
    call_tree = trace.get_calltree()
    assert call_tree.call_type == CallType.CALL
    assert repr(trace) == "__ETH_transfer__.0x() 1"


def test_request_timeout(connected_provider, project):
    # Test value set in `ape-config.yaml`
    expected = 29
    actual = connected_provider.web3.provider._request_kwargs["timeout"]
    assert actual == expected

    # Test default behavior
    with project.temp_config(foundry={"timeout": 30}):
        assert connected_provider.timeout == 30


def test_contract_interaction(connected_provider, owner, contract_instance, mocker):
    # Spy on the estimate_gas RPC method.
    estimate_gas_spy = mocker.spy(connected_provider.web3.eth, "estimate_gas")

    # Check what max gas is before transacting.
    max_gas = connected_provider.max_gas

    # Invoke a method from a contract via transacting.
    receipt = contract_instance.setNumber(102, sender=owner)

    # Verify values from the receipt.
    assert not receipt.failed
    assert receipt.receiver == contract_instance.address
    assert receipt.gas_used < receipt.gas_limit
    assert receipt.gas_limit == max_gas

    # Show contract state changed.
    assert contract_instance.myNumber() == 102

    # Verify the estimate gas RPC was not used (since we are using max_gas).
    assert estimate_gas_spy.call_count == 0


def test_revert(sender, contract_instance):
    # 'sender' is not the owner so it will revert (with a message)
    with pytest.raises(ContractLogicError, match="!authorized"):
        contract_instance.setNumber(6, sender=sender)

    # Show it also works with the reverts-context-manager.
    with reverts("!authorized"):
        contract_instance.setNumber(55, sender=sender)


def test_contract_revert_no_message(owner, contract_instance):
    # The Contract raises empty revert when setting number to 5.
    with pytest.raises(ContractLogicError, match="Transaction failed."):
        contract_instance.setNumber(5, sender=owner)


def test_transaction_contract_as_sender(
    contract_instance, contract_container, connected_provider, owner
):
    # Set balance so test wouldn't normally fail from lack of funds
    sender = owner.deploy(contract_container)
    connected_provider.set_balance(sender.address, "1000 ETH")

    with pytest.raises(ContractLogicError, match="!authorized"):
        # Task failed successfully
        contract_instance.setNumber(10, sender=sender)


@pytest.mark.parametrize(
    "amount", ("50 ETH", int(50e18), "0x2b5e3af16b1880000", "50000000000000000000")
)
def test_set_balance(connected_provider, owner, convert, amount):
    connected_provider.set_balance(owner.address, amount)
    assert owner.balance == convert("50 ETH", int)


def test_set_code(connected_provider, contract_container, owner):
    contract = contract_container.deploy(sender=owner)
    provider = connected_provider
    code = provider.get_code(contract.address)
    assert type(code) is HexBytes
    assert provider.set_code(contract.address, "0x00") is True
    assert provider.get_code(contract.address) != code
    assert provider.set_code(contract.address, code) is True
    assert provider.get_code(contract.address) == code


def test_set_storage(connected_provider, contract_container, owner):
    contract = contract_container.deploy(sender=owner)
    assert to_int(connected_provider.get_storage(contract.address, "0x2b5e3af16b1880000")) == 0
    connected_provider.set_storage(contract.address, "0x2b5e3af16b1880000", "0x1")
    assert to_int(connected_provider.get_storage(contract.address, "0x2b5e3af16b1880000")) == 1


def test_return_value(connected_provider, contract_instance, owner):
    tx = contract_instance.setAddress(owner, sender=owner)
    actual = tx.return_value
    expected = 123
    assert actual == expected


def test_return_value_tx_with_subcalls(connected_provider, contract_a, owner):
    tx = contract_a.methodWithoutArguments(sender=owner)
    actual = tx.return_value
    expected = HexBytes("0x0000000000000000000000003c44cdddb6a900fa2b585dd299e03d12fa4293bc")
    assert actual == expected


def test_get_receipt(connected_provider, contract_instance, owner):
    receipt = contract_instance.setAddress(owner.address, sender=owner)
    actual = connected_provider.get_receipt(receipt.txn_hash)
    assert receipt.txn_hash == actual.txn_hash
    assert actual.receiver == contract_instance.address
    assert actual.sender == receipt.sender


def test_revert_error(error_contract, not_owner):
    """
    Test matching a revert custom Solidity error.
    """
    with pytest.raises(error_contract.Unauthorized):
        error_contract.withdraw(sender=not_owner)


def test_revert_error_using_impersonated_account(error_contract, accounts, connected_provider):
    """
    Show that when a failure occurs when a txn is sent by an impersonated
    account, that everything still works.
    """
    acct = accounts[TEST_WALLET_ADDRESS]
    acct.balance = "1000 ETH"
    with pytest.raises(error_contract.Unauthorized):
        error_contract.withdraw(sender=acct)

    # Show we can "allow" reverts using impersonated accounts.
    # NOTE: This is extra because of the lack of tx-hash available.
    receipt = error_contract.withdraw(sender=acct, raise_on_revert=False)
    assert receipt.failed


def test_revert_allow(error_contract, not_owner, contract_instance):
    # 'sender' is not the owner so it will revert (with a message)
    receipt = error_contract.withdraw(sender=not_owner, raise_on_revert=False)
    assert receipt.error is not None
    assert isinstance(receipt.error, error_contract.Unauthorized)

    # Ensure this also works for calls.
    contract_instance.setNumber.call(5, raise_on_revert=False)


@pytest.mark.parametrize("host", ("https://example.com", "example.com"))
def test_host(project, local_network, host):
    with project.temp_config(foundry={"host": host}):
        provider = local_network.get_provider("foundry")
        assert provider.uri == "https://example.com"


def test_base_fee(connected_provider, project, networks, accounts):
    assert connected_provider.base_fee == 0

    acct1 = accounts[-1]
    acct2 = accounts[-2]

    # Show we can se the base-fee.
    new_base_fee = 1_000_000
    data = {"base_fee": new_base_fee, "host": "http://127.0.0.1:8555"}
    with project.temp_config(foundry=data):
        with networks.ethereum.local.use_provider("foundry") as provider:
            # Verify the block has the right base fee
            block_one = provider.get_block("latest")
            assert block_one.base_fee == new_base_fee

            # Make sure the command has the right base fee
            cmd = provider.build_command()
            idx = -1
            for i, part in enumerate(cmd):
                if part == "--block-base-fee-per-gas":
                    idx = i + 1
            assert idx > -1  # option was found
            assert cmd[idx] == str(new_base_fee)  # option val is correct

            # Show can transact with this base_fee
            acct1.transfer(acct2, "1 eth")

            # Verify the block still has the right base fee
            block_two = provider.get_block("latest")
            assert block_two.base_fee == new_base_fee


def test_auto_mine(connected_provider):
    assert connected_provider.auto_mine is True
    connected_provider.auto_mine = False
    assert connected_provider.auto_mine is False
    connected_provider.auto_mine = True
    assert connected_provider.auto_mine is True


@pytest.mark.parametrize(
    "message",
    (
        "Error: VM Exception while processing transaction: reverted with reason string ",
        "Transaction reverted without a reason string",
        "execution reverted",
    ),
)
def test_get_virtual_machine_error_from_contract_logic_message_includes_base_err(
    message, connected_provider
):
    exception = Exception(message)
    actual = connected_provider.get_virtual_machine_error(exception)
    assert isinstance(actual, ContractLogicError)
    assert actual.base_err == exception


def test_get_virtual_machine_error_first_arg_not_message(connected_provider):
    # Handling weird edge case I have not seen in the wild but was raised in a GH issue.
    exception = Exception(Exception())
    actual = connected_provider.get_virtual_machine_error(exception)
    assert isinstance(actual, VirtualMachineError)


def test_no_mining(project, local_network, connected_provider):
    assert "--no-mining" not in connected_provider.build_command()
    with project.temp_config(foundry={"auto_mine": "false"}):
        provider = local_network.get_provider("foundry")
        cmd = provider.build_command()
        assert "--no-mining" in cmd


def test_block_time(project, local_network, connected_provider):
    assert "--block-time" not in connected_provider.build_command()
    with project.temp_config(foundry={"block_time": 10}):
        provider = local_network.get_provider("foundry")
        cmd = provider.build_command()
        assert "--block-time" in cmd
        assert "10" in cmd


def test_remote_host(project, local_network, no_anvil_bin):
    with project.temp_config(foundry={"host": "https://example.com"}):
        with pytest.raises(
            FoundryProviderError,
            match=r"Failed to connect to remote Anvil node at 'https://example.com'\.",
        ):
            with local_network.use_provider("foundry"):
                assert True


def test_remote_host_using_env_var(local_network, no_anvil_bin):
    original = os.environ.get("APE_FOUNDRY_HOST")
    os.environ["APE_FOUNDRY_HOST"] = "https://example2.com"

    try:
        with pytest.raises(
            FoundryProviderError,
            match=r"Failed to connect to remote Anvil node at 'https://example2.com'\.",
        ):
            with local_network.use_provider("foundry") as provider:
                # It shouldn't actually get to the line below,
                # but in case it does, this is a helpful debug line.
                assert provider.uri == os.environ["APE_FOUNDRY_HOST"], "env var not setting."

    finally:
        if original is None:
            del os.environ["APE_FOUNDRY_HOST"]
        else:
            os.environ["APE_FOUNDRY_HOST"] = original


def test_send_transaction_when_no_error_and_receipt_fails(
    mocker, connected_provider, owner, vyper_contract_instance
):
    mock_web3 = mocker.MagicMock()
    mock_transaction = mocker.MagicMock()
    mock_transaction.required_confirmations = 0
    mock_transaction.sender = owner.address

    start_web3 = connected_provider._web3
    connected_provider._web3 = mock_web3

    try:
        # NOTE: Value is meaningless.
        tx_hash = HexBytes32.__eth_pydantic_validate__(123**36)

        # Sending tx "works" meaning no vm error.
        mock_web3.eth.send_raw_transaction.return_value = tx_hash

        # Getting a receipt "works", but you get a failed one.
        receipt_data = {
            "failed": True,
            "blockNumber": 0,
            "txnHash": to_hex(tx_hash),
            "status": TransactionStatusEnum.FAILING.value,
            "sender": owner.address,
            "receiver": vyper_contract_instance.address,
            "input": b"",
            "gasUsed": 123,
            "gasLimit": 100,
        }
        mock_web3.eth.wait_for_transaction_receipt.return_value = receipt_data

        # Attempting to replay the tx does not produce any error.
        mock_web3.eth.call.return_value = HexBytes("")

        # Execute test.
        with pytest.raises(TransactionError):
            connected_provider.send_transaction(mock_transaction)

    finally:
        connected_provider._web3 = start_web3


@pytest.mark.parametrize("tx_type", TransactionType)
def test_prepare_tx_with_max_gas(tx_type, connected_provider, ethereum, owner):
    tx = ethereum.create_transaction(type=tx_type.value, sender=owner.address)
    tx.gas_limit = None  # Undo set from validator
    assert tx.gas_limit is None, "Test setup failed - couldn't clear tx gas limit."

    # NOTE: The local network by default uses max_gas.
    actual = connected_provider.prepare_transaction(tx)
    assert actual.gas_limit == connected_provider.max_gas


def test_disable_block_gas_limit(project, disconnected_provider):
    # Ensure it is disabled by default.
    cmd = disconnected_provider.build_command()
    assert "--disable-block-gas-limit" not in cmd

    # Show we can enable it.
    with project.temp_config(foundry={"disable_block_gas_limit": True}):
        cmd = disconnected_provider.build_command()
        assert "--disable-block-gas-limit" in cmd


def test_initial_balance(accounts):
    # The value is set in the config, but checking it can be less just in case.
    # The regular default value is 10_000 so hopefully it isn't below that,
    # just showing we were able to increase it.
    acct = accounts[9]
    assert convert("10_000 ETH", int) < acct.balance <= convert("100_000 ETH", int)


@pytest.mark.parametrize("host", ("https://example.com", "example.com"))
def test_evm_version(project, local_network, host):
    with project.temp_config(foundry={"evm_version": "shanghai"}):
        provider = local_network.get_provider("foundry")
        cmd = provider.build_command()
        assert "--hardfork" in cmd
        assert "shanghai" in cmd


def test_optimism(networks):
    with networks.optimism.local.use_provider(
        "foundry", provider_settings={"port": 9545}
    ) as provider:
        assert provider.is_connected
        cmd = provider.build_command()
        assert "--optimism" in cmd



from functools import cached_property
from typing import Any

from ape.exceptions import ContractNotFoundError
from ape_ethereum.trace import TraceApproach, TransactionTrace
from hexbytes import HexBytes


class AnvilTransactionTrace(TransactionTrace):
    call_trace_approach: TraceApproach = TraceApproach.PARITY
    debug_trace_transaction_parameters: dict = {
        "stepsTracing": True,
        "enableMemory": True,
    }

    @cached_property
    def return_value(self) -> Any:
        if self._enriched_calltree:
            # Only check enrichment output if was already enriched!
            # Don't enrich ONLY for return value, as that is very bad performance
            # for realistic contract interactions.
            return self._return_value_from_enriched_calltree

        # perf: Avoid any model serializing/deserializing that happens at
        #   Ape's abstract layer at this point.
        trace_tx_iter = self.provider.stream_request("trace_transaction", [self.transaction_hash])
        if not (top_level_call := next(trace_tx_iter, None)):
            return (None,)

        try:
            address = top_level_call["action"]["to"]
            calldata = top_level_call["action"]["input"]
            contract_type = self.chain_manager.contracts[address]
            abi = contract_type.methods[calldata[:10]]
        except (KeyError, ContractNotFoundError):
            abi = self.root_method_abi

        if output := top_level_call.get("result", {}).get("output"):
            return self._ecosystem.decode_returndata(abi, HexBytes(output))

        return (None,)



from ape.api import ReceiptAPI


def test_contract_transaction_revert(benchmark, connected_provider, owner, contract_instance):
    tx = benchmark.pedantic(
        lambda *args, **kwargs: contract_instance.setNumber(*args, **kwargs),
        args=(5,),
        kwargs={"sender": owner, "raise_on_revert": False},
        rounds=5,
        warmup_rounds=1,
    )
    assert isinstance(tx, ReceiptAPI)  # Sanity check.
    stats = benchmark.stats
    median = stats.get("median")

    # Was seeing 0.44419266798649915.
    # Seeing 0.2634877339878585 as of https://github.com/ApeWorX/ape-foundry/pull/115
    assert median < 3.5



# Extracted from Erigon chain specs
# https://github.com/ledgerwatch/erigon/tree/devel/params/chainspecs
# https://gist.github.com/banteg/e38c7d98f35286c303311168f28404ca
EVM_VERSION_BY_NETWORK = {
    "ethereum": {
        "mainnet": {
            0: "frontier",
            1150000: "homestead",
            4370000: "byzantium",
            7280000: "petersburg",
            9069000: "istanbul",
            9200000: "muirglacier",
            12244000: "berlin",
            12965000: "london",
        },
        "sepolia": {
            0: "london",
        },
        "holesky": {
            0: "london",
        },
    },
    "polygon": {
        "mainnet": {
            0: "petersburg",
            3395000: "muirglacier",
            14750000: "berlin",
            23850000: "london",
        },
        "amoy": {
            0: "berlin",
            73100: "london",
        },
    },
}
